pub struct CollectionStats {
pub name: String,
pub config: CollectionConfig,
pub live: usize,
pub deleted: usize,
pub nodes_per_layer: Vec<usize>,
pub avg_degree_layer0: f64,
pub unreachable: usize,
pub vector_bytes: usize,
pub graph_bytes: usize,
pub metadata_bytes: usize,
}Fields
name: Stringconfig: CollectionConfiglive: usizedeleted: usizeDeleted or replaced records still occupying space until compact.
nodes_per_layer: Vec<usize>Nodes on each HNSW layer, bottom first.
avg_degree_layer0: f64unreachable: usizeLive records that graph search can never reach from the entry point.
vector_bytes: usizegraph_bytes: usizemetadata_bytes: usizeTrait Implementations
Sourceimpl Clone for CollectionStats
impl Clone for CollectionStats
Sourceimpl Debug for CollectionStats
impl Debug for CollectionStats
Sourceimpl PartialEq for CollectionStats
impl PartialEq for CollectionStats
impl StructuralPartialEq for CollectionStats